// Memory debugging classes implementation
//
// Copyright (c) David Drysdale 2001
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.

#define TRCCLASSNAME "DebugMem"
#include <cpptrace.h>
#ifdef MT_DEBUGMEM
#include <lock.h>
#endif

// Overview:
// Each piece of memory allocated gets some extra bits connected to it:
//
//   [Size] [ID] <memory used> [Guard bytes]
//               ^
//               |
//               Pointer returned

#include <debugmem.h>
#include <cstdlib>
#include <cstring>
#include <new>
#include <map>
#include <string>

//! Number of guard bytes
static const int NUMGUARDS = 4;
//! Byte used to fill uninitialized memory
static const unsigned char FILLBYTE=0xCD;
//! Byte value used in guard bytes
static const unsigned char GUARDBYTE=0xEF;
//! Marker to indicate that memory is valid (has been allocated and not freed)
static const long VALIDID=0x0ADAFAFA;

//! Pair holding filename and line number information
typedef std::pair<const char*, int> CodeLocation;

// Single instance class to check leaks at end of day
class MemoryChecker {
private:
  //! Map from memory pointer to code location that allocated the memory
  std::map<const void*, CodeLocation> blocks;
public:
  //! Check whether a given memory pointer is known about
  bool checkEntry(const void *p) const {return(blocks.find(p)!=blocks.end());}
  //! Add entry to tracked collection of memory blocks
  void addEntry(const void *p, const char* f, int l);
  //! Remove entry from tracked collection of memory blocks
  void removeEntry(const void *p);
  //! Constructor
  MemoryChecker() {}
  //! Destructor; emits errors for leaked blocks
  ~MemoryChecker();
};
void MemoryChecker::addEntry(const void *p, const char* f, int l)
{
  TRACE_FN("addEntry");
  std::map<const void*, CodeLocation>::iterator pos = blocks.find(p);
  if (pos != blocks.end()) {
    TRC_ERR("already got an entry for block at "<<p);
    TRC_ERR("previously seen from "<<pos->second.first<<":"<<pos->second.second);
  }
  blocks[p].first = f;
  blocks[p].second = l;
}
void MemoryChecker::removeEntry(const void *p)
{
  TRACE_FN("removeEntry");
  std::map<const void*, CodeLocation>::iterator pos = blocks.find(p);
  if (pos==blocks.end()) {
    TRC_ERR("no entry to remove for "<<p);
  } else {
    blocks.erase(pos);
  }
}
MemoryChecker::~MemoryChecker()
{
  for (std::map<const void*, CodeLocation>::iterator ii = blocks.begin();
       ii != blocks.end(); ii++) {
    const void* p = ii->first;
    CodeLocation w=ii->second;
    std::cerr<<"block at "<<p<<" leaked; originally allocated at "<<w.first<<":"<<w.second<<std::endl;
  }
}
//! Global instance of the memory block tracker
static MemoryChecker sMemList;
#ifdef MT_DEBUGMEM
//! Lock protecting access
DECLARE_LOCK(sMemListLock);
#define WITH_LOCKED_MEMLIST() BLOCK_LOCK(sMemListLock)
#else
#define WITH_LOCKED_MEMLIST()
#endif

//! Allocation function
void* debug_new(size_t size, const char *file, int line) throw (std::bad_alloc)
{
  TRACE_FN("debug_new");
  unsigned char* ptr;

  TRC_DBG("asked for "<<size<<" bytes by "<<file<<":"<<line);
  // Enforce minimum size
  if (size == 0) size=1;
  while (true) {
    // Get raw memory via malloc().  Allow extra size for additional information:
    //   [Size] [ID] <memory used> [Guard bytes]
    // This could theoretically cause alignment issues, but in practice should not because
    // we are preprending a size_t and a long -- on 32-bit systems this will be 8 bytes,
    // and on 64-bit LP64 systems this will be a 16 bytes.
    ptr = static_cast<unsigned char*>(malloc(sizeof(size_t)+sizeof(long)+size+NUMGUARDS));
    if (ptr) {
      // First, put in the size of the memory
      size_t *pS = reinterpret_cast<size_t*>(ptr);
      *pS = size;
      ptr += sizeof(size_t);
      // Next, put in a validity ID field
      long *pI = reinterpret_cast<long*>(ptr);
      *pI = VALIDID;
      ptr += sizeof(long);
      // Next, pre-initialize the memory
      memset(ptr, FILLBYTE, size);
      // Finally, put in some guard bytes at the end
      memset(ptr+size, GUARDBYTE, NUMGUARDS);
      // Track the returned block of memory
      void *p = static_cast<void *>(ptr);
      TRC_DBG("returning "<<p);
      WITH_LOCKED_MEMLIST();
      sMemList.addEntry(p, file, line);
      return p;
    } else {
      // Allocation was unsuccessful; find out what the current
      // error handling function is
      std::new_handler gH = std::set_new_handler(0);
      std::set_new_handler(gH);

      if (gH) (*gH)(); else throw std::bad_alloc();
    }
  }
}

void debug_delete(const void *p, const char* file, int line) throw ()
{
  TRACE_FN("debug_delete");
  TRC_DBG("asked to delete memory at "<<p<<" by "<<file<<":"<<line);

  // Deleting null is allowed
  if (p==0) return;

  // Move back in memory to find the size field
  const unsigned char *ptr = static_cast<const unsigned char *>(p);
  const size_t *pS = reinterpret_cast<const size_t*>(ptr - sizeof(long) - sizeof(size_t));
  size_t size = *pS;

  // Check then trash the ID field
  long *pI = reinterpret_cast<long*>(const_cast<unsigned char*>(ptr - sizeof(long)));
  if (*pI != VALIDID) {
    // looks like it has already been freed
    TRC_ERR(file<<":"<<line<<": unexpected attempt to free memory at "<<p<<" - already freed?");
    return;
  }
  *pI = -1;

  // Check the guard bytes are intact
  const unsigned char *pG = ptr + size;
  for (int ii=0; ii<NUMGUARDS; ii++) {
    if (pG[ii] != GUARDBYTE) {
      TRC_ERR(file<<":"<<line<<": guard byte "<<ii<<" overwritten with "<< static_cast<int>(pG[ii])<<
              ": in block at "<<p<<" size "<<size);
    }
  }

  // Remove from the master tracker
  WITH_LOCKED_MEMLIST();
  sMemList.removeEntry(p);

  // Return the raw memory
  free(const_cast<size_t *>(pS));
}

//! Placement new that takes 2 additional arguments -- filename and line number
//! Used for calls like:
//!   Thing* t = new("fred.c", 213) Thing();
void* operator new(size_t size, const char *file, int line) throw (std::bad_alloc)
{
  return debug_new(size, file, line);
}
//! Matching placement delete.  This is only called if a constructor throws
//! just after placement new has been used.
void operator delete(void* ptr, const char *file, int line) throw ()
{
  debug_delete(ptr, file, line);
}
