// -*-c++-*-
#ifndef _LOCK_H_
#define _LOCK_H_
// Lock objects for UNIX and Win32
//
// Copyright (c) David Drysdale 2000,2010
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.
// --------------------------------------------------------------------------
//
// To use these, you need to define NO_PTHREADS if you want to get the native
// windows implementation.  If DEBUG is defined, then additional information
// is emitted to the trace file, and attempts are made to spot potential
// deadlocks.
//
// Use the following macros:
//   DECLARE_LOCK(sLockName);   // to declare a lock, at file or block scope
//   BLOCK_LOCK(sLockName);     // lock within a block scope
//   LOCK(sLockName);           // to enter a lock, which will block if another thread
//                              // already has the lock
//   UNLOCK(sLockName);         // to release a lock
//
// BLOCK_LOCK is preferred, since it is exception-safe.
//
// Putting this together, a sample file using this locking might look like:
//
//  #include <lock.h>
//
//  DECLARE_LOCK(sThingLock);
//  int thing;
//  void f(int a)
//  {
//    LOCK(sThingLock);
//    if (thing == 0) thing=a;
//    UNLOCK(sThingLock);
//    if (other) {
//      BLOCK_LOCK(sThingLock);
//      do stuff, possibly throwing exceptions
//    }
//  }
//
// NOTE: you cannot nest locks, i.e. call LOCK twice for the same lock in the same thread
//
// Finally, at link time you'll need to include
//  a) the lock.o object, created from lock.cpp
//  b) the standard threads library, -lpthread for POSIX threads

#ifndef NO_PTHREADS
// POSIX threads version of mutex/critical section code
#include <pthread.h>
#else
// Windows version of mutex/critical section code
#include <winbase.h>
typedef CriticalSection pthread_mutex_t;
#define pthread_mutex_init(pS, ATTR) InitializeCriticalSection(pS)
#define pthread_mutex_destroy(pS)    DeleteCriticalSection(pS)
#define pthread_mutex_lock(pS)       EnterCriticalSection(pS)
#define pthread_mutex_unlock(pS)     LeaveCriticalSection(pS)
#endif

// Normal definition of an object which encapsulates a mutex/critical section
// The constructor/destructor take care of init/term
// NOTE: calls to getLock()/releaseLock from the same thread may NOT be nested
// with this object (since the POSIX implementation will not support them)
class CriticalSectionObject {
 public:
  CriticalSectionObject()          {pthread_mutex_init(&mMutex, NULL);}
  virtual ~CriticalSectionObject() {pthread_mutex_destroy(&mMutex);}
  void getLock()                   {pthread_mutex_lock(&mMutex);}
  void releaseLock()               {pthread_mutex_unlock(&mMutex);}
 private:
  pthread_mutex_t mMutex;
};

class CSOAcquisition {
 public:
  CSOAcquisition(CriticalSectionObject& cso) : mCso(cso) {mCso.getLock();}
  ~CSOAcquisition() {mCso.releaseLock();}
 private:
  CriticalSectionObject& mCso;
};

#ifndef DEBUG
#define DECLARE_LOCK(NAME) CriticalSectionObject NAME
#define BLOCK_LOCK(NAME)   CSOAcquisition acquired(NAME)
#define LOCK(NAME)         NAME.getLock()
#define UNLOCK(NAME)       NAME.releaseLock()
#else
#define DECLARE_LOCK(NAME) DebugCriticalSectionObject NAME(#NAME,__FILE__,__LINE__)
#define BLOCK_LOCK(NAME)   DebugCSOAcquisition acquired(NAME,__FILE__,__LINE__)
#define LOCK(NAME)         NAME.getLock(__FILE__,__LINE__)
#define UNLOCK(NAME)       NAME.releaseLock(__FILE__,__LINE__)

#include <string>

// This is the debug version of the critical section object.  It performs tracing
// and checking for potential deadlocks
class DebugCriticalSectionObject : public CriticalSectionObject{
public:
  DebugCriticalSectionObject(const char* name, const char* filename, int line);
  virtual ~DebugCriticalSectionObject();
  void getLock(const char* filename, int line);
  void releaseLock(const char* filename, int line);
private:
  std::string mLabel;
  int mLockCount;
};
class DebugCSOAcquisition {
 public:
  DebugCSOAcquisition(DebugCriticalSectionObject& dcso, const char* filename, int line) :
    mDcso(dcso) {mDcso.getLock(filename, line);}
  ~DebugCSOAcquisition() {mDcso.releaseLock("<end of block>", 0);}
 private:
  DebugCriticalSectionObject& mDcso;
};

#endif

#endif
