// C++ Trace Control Program
//
// Copyright (c) David Drysdale 2000,2010
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.

// No point in building a non-trace-enabled version of the trace control program,
// so force definition of TRC_COMPILE_LEVEL
#ifdef TRC_COMPILE_LEVEL
#undef TRC_COMPILE_LEVEL
#endif
#define TRC_COMPILE_LEVEL 0
#include <cpptrace.h>

#include <getopt_local.h>
#include <iostream>

void printusage()
{
  std::cerr << "Trace Control Program"<<std::endl<<std::endl;
  std::cerr << " -h     - Help (this output)"<<std::endl;
  std::cerr << " -q     - Query current configuration"<<std::endl;
  std::cerr << " -L n   - Set overall trace level"<<std::endl;
  std::cerr << " -S n   - Set max trace file size"<<std::endl;
  std::cerr << " -t n   - Turn thread ID tracing on(1) or off(0)"<<std::endl;
  std::cerr << " -r     - Reset trace files"<<std::endl;
  std::cerr << " -c n   - Set field width for class name"<<std::endl;
  std::cerr << " -m n   - Set field width for method name"<<std::endl;
  std::cerr << " -n n   - Set field width for line number"<<std::endl;
}

int main(int argc, char *argv[])
{
  bool doQuery=false;
  long newLevel=MIN_TRACE_LEVEL-1;
  int newSize=-1;
  int newClassWidth=-1;
  int newMethodWidth=-1;
  int newLineWidth=-1;
  int newTraceTID=-1;
  bool doReset=false;
  int opt;

  while ((opt = getopt_l(argc, argv,"hqL:S:rc:n:m:t:")) != -1) {
    switch (opt) {
      case 'h': printusage(); break;
      case 'q': doQuery=true; break;
      case 'L': newLevel=atol(optarg_l); break;
      case 'S': newSize=atoi(optarg_l); break;
      case 'r': doReset=true; break;
      case 'c': newClassWidth=atoi(optarg_l); break;
      case 'm': newMethodWidth=atoi(optarg_l); break;
      case 'n': newLineWidth=atoi(optarg_l); break;
      case 't': newTraceTID=atoi(optarg_l); break;
      default: printusage(); break;
    }
  }
  argc -= optind_l;
  argv += optind_l;

  // Connect to the trace information, and lock it (until we exit the scope)
  TraceOverall* pT = TraceOverall::instance();
  TraceLock scopeLock(TraceOverall::sLockFD);

  // Now perform all of the requested actions
  if (newLevel >= MIN_TRACE_LEVEL) {
    std::cout << "Set trace level to " << newLevel << std::endl;
    pT->setTraceLevel(newLevel);
  }
  if (newSize > 0) {
    std::cout << "Set max trace file size to " << newSize << std::endl;
    pT->setMaxFileSize(newSize);
  }
  if (newClassWidth >= 0) {
    std::cout << "Set class name display width to " << newClassWidth << std::endl;
    pT->setClassWidth(newClassWidth);
  }
  if (newMethodWidth >= 0) {
    std::cout << "Set method name display width to " << newMethodWidth << std::endl;
    pT->setMethodWidth(newMethodWidth);
  }
  if (newLineWidth >= 0) {
    std::cout << "Set line number display width to " << newLineWidth << std::endl;
    pT->setLineWidth(newLineWidth);
  }
  if (newTraceTID >= 0) {
    std::cout << "Set thread ID tracing to " << (newTraceTID !=0) << std::endl;
    pT->setIncludeTID((newTraceTID != 0));
  }
  if (doReset) {
    std::cout << "Reset trace files" << std::endl;
    delete new std::ofstream(pT->getTraceFile1(), std::ios::trunc|std::ios::out);
    delete new std::ofstream(pT->getTraceFile2(), std::ios::trunc|std::ios::out);
  }
  if (doQuery) {
    std::cout<<"Tracing Configuration (version "<<pT->getMajorVersion()<<"."<<
        pT->getMinorVersion()<<")"<<std::endl;
    std::cout<<"Trace level             "<<pT->getTraceLevel()<<std::endl;
    std::cout<<"Include thread ID       "<<pT->getIncludeTID()<<std::endl;
    std::cout<<"Maximum file size       "<<pT->getMaxFileSize()<<std::endl;
    std::cout<<"Trace file 1            "<<pT->getTraceFile1()<<std::endl;
    std::cout<<"Trace file 2            "<<pT->getTraceFile2()<<std::endl;
    std::cout<<"Number of trace users   "<<(pT->getUseCount()-1)<<std::endl;
    std::cout<<"Class field width       "<<pT->getClassWidth()<<std::endl;
    std::cout<<"Method field width      "<<pT->getMethodWidth()<<std::endl;
    std::cout<<"Line number field width "<<pT->getLineWidth()<<std::endl;
  }
  return(0);
}
